# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Authentication methods
"""

from knack.util import CLIError

from paconn.authentication.profile import Profile
from paconn.authentication.tokenmanager import TokenManager


def get_authentication(settings, force_authenticate):
    """
    Logs the user in and saves the token in a file.
    """
    tokenmanager = TokenManager()
    credentials = tokenmanager.read()

    token_expired = TokenManager.is_expired(credentials)

    # Get new token
    if token_expired or force_authenticate:
        profile = Profile(
            client_id=settings.client_id,
            tenant=settings.tenant,
            resource=settings.resource,
            authority_url=settings.authority_url)

        credentials = profile.authenticate_device_code()

        tokenmanager.write(credentials)

        token_expired = TokenManager.is_expired(credentials)

    # Couldn't acquire valid token
    if token_expired:
        raise CLIError('Couldn\'t get authentication')



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Validate command.
"""

from paconn import _VALIDATE

from paconn.common.util import display
from paconn.settings.util import load_powerapps_and_flow_rp
from paconn.settings.settingsbuilder import SettingsBuilder

import paconn.operations.validate


def validate(
        api_definition,
        powerapps_url,
        powerapps_version,
        settings_file):
    """
    Validate command.
    """
    # Get settings
    settings = SettingsBuilder.get_settings(
        environment=None,
        settings_file=settings_file,
        api_properties=None,
        api_definition=api_definition,
        icon=None,
        connector_id=None,
        powerapps_url=powerapps_url,
        powerapps_version=powerapps_version)

    powerapps_rp, _ = load_powerapps_and_flow_rp(
        settings=settings,
        command_context=_VALIDATE)

    result = paconn.operations.validate.validate(
        powerapps_rp=powerapps_rp,
        settings=settings)

    if result:
        display(result)
    else:
        display('{} validated successfully.'.format(settings.api_definition))



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Flow API Manager
"""


# pylint: disable=too-few-public-methods
class FlowRP:
    """
    Flow API Manager class
    """
    def __init__(self, api_manager):
        self.api_manager = api_manager

    def get_environments(self):
        """
        Returns a list of environments
        """
        endpoint = self.api_manager.construct_url(
            path='environments')

        response = self.api_manager.request(
            verb='GET',
            endpoint=endpoint)

        return response.json()



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
A builder class to create a PowerAppsRP object
"""

from paconn.apimanager.apimanagerbuilder import APIManagerBuilder
from paconn.apimanager.flowrp import FlowRP


# pylint: disable=too-few-public-methods
class FlowRPBuilder:
    """
    A builder class to create a FlowRP object
    """
    def get_from_settings(credentials, settings):
        """
        Returns flow rp object from a given settings and credentials.
        """

        # Create the API Manager
        flow_api_manager = APIManagerBuilder.get_from_url(
            url=settings.flow_url,
            base_path=settings.flow_base_path,
            api_version=settings.flow_api_version,
            credentials=credentials)

        flow_rp = FlowRP(api_manager=flow_api_manager)

        return flow_rp



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Defines main entry point for the CLI
"""

from __future__ import print_function
import sys

from knack import CLI, CLICommandsLoader
# pylint: disable=unused-import
from paconn.commands.help import helps  # noqa: F401
from paconn.common.util import get_config_dir


class ConnectorsCli(CLI):
    """
    The CLI CLass
    """
    def get_cli_version(self):
        """
        Returns cli version
        """
        from . import __VERSION__
        return __VERSION__


class ConnectorsCliCommandsLoader(CLICommandsLoader):
    """
    Commands loader
    """
    def load_command_table(self, args):
        """
        Loads command table
        """
        from paconn.commands.commands import load_command_table

        load_command_table(self, args)
        return super(ConnectorsCliCommandsLoader, self).load_command_table(args)

    def load_arguments(self, command):
        """
        Load arguments
        """
        from paconn.commands.params import load_arguments

        load_arguments(self, command)
        super(ConnectorsCliCommandsLoader, self).load_arguments(command)


def main():
    """
    The main method
    """

    try:
        from paconn import __CLI_NAME__
        cli_context = ConnectorsCli(
            cli_name=__CLI_NAME__,
            commands_loader_cls=ConnectorsCliCommandsLoader,
            config_dir=get_config_dir())

        exit_code = cli_context.invoke(sys.argv[1:])
        sys.exit(exit_code)
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == '__main__':
    main()



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Help strings for the CLI
"""

from knack.help_files import helps  # pylint: disable=unused-import
from paconn import _COMMAND_GROUP, _LOGIN, _DOWNLOAD, _CREATE, _UPDATE, _VALIDATE

helps[_COMMAND_GROUP] = """
    short-summary: Microsoft Power Platform Connectors CLI
"""

helps[_LOGIN] = """
    type: command
    short-summary: Login to Power Platform.
    examples:
        - name: Login
          text: paconn login
"""


helps[_DOWNLOAD] = """
    type: command
    short-summary: Downloads a given custom connector to the local directory.
    examples:
        - name: Download connector
          text: paconn download
"""

helps[_CREATE] = """
    type: command
    short-summary: Creates a new custom connector from the given directory.
    examples:
        - name: Create connector
          text: paconn create
"""

helps[_UPDATE] = """
    type: command
    short-summary: Update a given custom connector from the local directory.
    examples:
        - name: Update connector.
          text: paconn update
"""

helps[_VALIDATE] = """
    type: command
    short-summary: Validate the swagger for certification.
    examples:
        - name: Validate swagger
          text: paconn validate
"""



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Method for create/update operation
"""

import os
import json
import urllib.parse

from knack.util import CLIError

from paconn.common.util import ensure_file_exists
from paconn.settings.util import write_settings
from paconn.apimanager.iconuploader import upload_icon
from paconn.operations.json_keys import (
    _PROPERTIES,
    _ICON_URI,
    _OPEN_API_DEFINITION,
    _ENVIRONMENT,
    _NAME,
    _BACKEND_SERVICE,
    _SERVICE_URL,
    _SCHEMES,
    _HOST,
    _BASE_PATH,
    _DISPLAY_NAME,
    _CONNECTION_PARAMETERS,
    _TOKEN,
    _OAUTH_SETTINGS,
    _CLIENT_SECRET,
    _DESCRIPTION,
    _INFO,
    _TITLE,
    _SHARED_ACCESS_SIGNATURE
)


def _create_backendservice_url(openapi_definition):
    """
    Create a backend service URL from the Open API definition.
    """

    schemes = openapi_definition.get(_SCHEMES, [])
    scheme = next(iter(schemes), '')
    netloc = openapi_definition.get(_HOST, '')
    path = openapi_definition.get(_BASE_PATH, '')

    params = ''
    query = ''
    fragment = ''

    parts = (scheme, netloc, path, params, query, fragment)

    url = urllib.parse.urlunparse(parts)

    return url


def upsert(powerapps_rp, settings, client_secret, is_update, overwrite_settings):
    """
    Method for create/update operation
    """

    # Make sure the required files exist
    ensure_file_exists(
        file=settings.api_properties,
        file_type='API Properties')
    ensure_file_exists(
        file=settings.api_definition,
        file_type='API Definition')

    # Open the property file
    with open(settings.api_properties, 'r') as file:
        property_definition = json.load(file)

    # Get the property object
    properties = property_definition[_PROPERTIES]

    # Add secret
    token_property = properties.get(_CONNECTION_PARAMETERS, {}).get(_TOKEN, None)
    if token_property:
        oauth_settings = token_property.get(_OAUTH_SETTINGS, None)
        if oauth_settings and client_secret:
            oauth_settings[_CLIENT_SECRET] = client_secret
        elif oauth_settings and not client_secret and not is_update:
            raise CLIError('Please provide OAuth2 client secret using the --secret argument.')

    # Load swagger definition
    with open(settings.api_definition, 'r') as file:
        openapi_definition = json.load(file)

    # Append swagger
    properties[_OPEN_API_DEFINITION] = openapi_definition

    # Add backend service
    backend_service_url = _create_backendservice_url(openapi_definition)
    properties[_BACKEND_SERVICE] = {_SERVICE_URL: backend_service_url}

    # Append the environment id
    properties[_ENVIRONMENT] = {_NAME: settings.environment}

    # Add displayName only when creating a new connector
    if is_update is not True:
        properties[_DISPLAY_NAME] = openapi_definition[_INFO][_TITLE]

    # Add description
    properties[_DESCRIPTION] = openapi_definition[_INFO][_DESCRIPTION]

    # Validate Open API Definition
    powerapps_rp.validate_connector(
        payload=openapi_definition,
        enable_certification_rules=False)

    # Get the shared access signature
    response = powerapps_rp.generate_resource_storage(settings.environment)
    sas_url = response[_SHARED_ACCESS_SIGNATURE]

    # Upload the icon
    if settings.icon and os.path.exists(settings.icon):
        icon_uri = upload_icon(
            sas_url=sas_url,
            file_path=settings.icon)
        properties[_ICON_URI] = icon_uri

    # Update or create the connector
    if is_update is True:
        api_registration = powerapps_rp.update_connector(
            environment=settings.environment,
            connector_id=settings.connector_id,
            payload=property_definition)
        connector_id = settings.connector_id
    else:
        api_registration = powerapps_rp.create_connector(
            environment=settings.environment,
            payload=property_definition)
        connector_id = json.loads(api_registration)[_NAME]

        # Save the settings
        settings.connector_id = connector_id
        write_settings(settings, overwrite_settings)

    return connector_id



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Save operation.
"""

import os
import json
import requests

from knack.util import CLIError
from knack.prompting import prompt_y_n

from paconn.common.util import format_json
from paconn.settings.util import write_settings, SETTINGS_FILE

from paconn.operations.json_keys import (
    _PROPERTIES,
    _API_DEFINITIONS,
    _ORIGINAL_SWAGGER_URL,
    _ICON_URI,
    _CONNECTION_PARAMETERS,
    _ICON_BRAND_COLOR,
    _CAPABILITIES,
    _POLICY_TEMPLATE_INSTANCES
)


def _prepare_directory(destination, connector_id):
    """
    Create directory for saving a connector.
    """

    # Use the destination directory when provided
    if destination:
        if not os.path.exists(destination):
            # Create all sub-directories
            os.makedirs(destination)
    # Create a sub-directory in the current directory
    # when a destination isn't provided
    else:
        if os.path.isdir(connector_id):
            error = '{} directory already exists. Please remove the directory before continuing.'
            raise CLIError(error.format(connector_id))
        else:
            os.mkdir(connector_id)
            destination = connector_id

    if os.path.isdir(destination):
        os.chdir(destination)
    else:
        error = 'Couldn\'t download to the desination directory {}.'
        raise CLIError(error.format(destination))

    return os.getcwd()


def _ensure_overwrite(settings):
    """
    Ensure the files can be overwritten, if exists
    """
    overwrite = False
    files = [settings.api_properties, settings.api_definition, settings.icon, SETTINGS_FILE]
    existing_files = [file for file in files if os.path.exists(file)]
    if len(existing_files) > 0:
        msg = '{} file(s) exist. Do you want to overwrite?'.format(existing_files)
        overwrite = prompt_y_n(msg)
        if not overwrite:
            raise CLIError('{} files not overwritten.'.format(existing_files))

    return overwrite


def download(powerapps_rp, settings, destination, overwrite):
    """
    Download operation.
    """
    # Prepare folders
    directory = _prepare_directory(
        destination=destination,
        connector_id=settings.connector_id)

    # Check if files could be overwritten
    if not overwrite:
        overwrite = _ensure_overwrite(settings)

    api_registration = powerapps_rp.get_connector(
        environment=settings.environment,
        connector_id=settings.connector_id)

    if _PROPERTIES not in api_registration:
        raise CLIError('Properties not present in the api registration information.')

    api_properties = api_registration[_PROPERTIES]

    # Save the settings
    write_settings(settings, overwrite)

    # Property whitelist
    property_keys_whitelist = [
        _CONNECTION_PARAMETERS,
        _ICON_BRAND_COLOR,
        _CAPABILITIES,
        _POLICY_TEMPLATE_INSTANCES
    ]

    # Remove the keys that aren't present in the property JSON
    properties_present = list(
        filter(lambda prop: prop in api_properties, property_keys_whitelist)
    )

    # Only output the white listed properties that are present in the property JSON
    api_properties_selected = {_PROPERTIES: {}}
    api_properties_selected[_PROPERTIES] = {
        prop: api_properties[prop]
        for prop in properties_present
    }

    # Write the api properties
    api_prop = format_json(
        content=api_properties_selected,
        sort_keys=False)

    open(
        file=settings.api_properties,
        mode='w'
        ).write(api_prop)

    # Write the open api definition,
    # either from swagger URL when available or from swagger property.
    if _API_DEFINITIONS in api_properties and _ORIGINAL_SWAGGER_URL in api_properties[_API_DEFINITIONS]:
        original_swagger_url = api_properties[_API_DEFINITIONS][_ORIGINAL_SWAGGER_URL]
        response = requests.get(original_swagger_url, allow_redirects=True)
        response_string = response.content.decode('utf-8-sig')

        swagger = format_json(
            content=json.loads(response_string),
            sort_keys=False)

        open(
            file=settings.api_definition,
            mode='w'
            ).write(swagger)

    # Write the icon
    if _ICON_URI in api_properties:
        icon_url = api_properties[_ICON_URI]
        response = requests.get(icon_url, allow_redirects=True)

        open(
            file=settings.icon,
            mode='wb'
            ).write(response.content)

    return directory



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Create command.
"""

from paconn import _CREATE
from paconn.common.util import display
from paconn.settings.util import load_powerapps_and_flow_rp
from paconn.operations.upsert import upsert
from paconn.settings.settingsbuilder import SettingsBuilder


# pylint: disable=too-many-arguments
def create(
        environment,
        api_properties,
        api_definition,
        icon,
        powerapps_url,
        powerapps_version,
        client_secret,
        settings_file,
        overwrite_settings):
    """
    Create command.
    """
    # Get settings
    settings = SettingsBuilder.get_settings(
        environment=environment,
        settings_file=settings_file,
        api_properties=api_properties,
        api_definition=api_definition,
        icon=icon,
        connector_id=None,
        powerapps_url=powerapps_url,
        powerapps_version=powerapps_version)

    powerapps_rp, _ = load_powerapps_and_flow_rp(
        settings=settings,
        command_context=_CREATE)

    connector_id = upsert(
        powerapps_rp=powerapps_rp,
        settings=settings,
        client_secret=client_secret,
        is_update=False,
        overwrite_settings=overwrite_settings)

    display('{} created successfully.'.format(connector_id))



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Defines the command table
"""

from knack.commands import CommandGroup

from paconn import __CLI_NAME__
from paconn import _COMMAND_GROUP, _LOGIN, _DOWNLOAD, _CREATE, _UPDATE, _VALIDATE


# pylint: disable=unused-argument
def load_command_table(self, args):
    """
    Loads the command table
    """
    def operation_group(name):
        return '{cli_name}.commands.{name}#{name}'.format(cli_name=__CLI_NAME__, name=name)

    with CommandGroup(self, _COMMAND_GROUP, operation_group(_LOGIN)) as command_group:
        command_group.command(_LOGIN, _LOGIN)

    with CommandGroup(self, _COMMAND_GROUP, operation_group(_DOWNLOAD)) as command_group:
        command_group.command(_DOWNLOAD, _DOWNLOAD)

    with CommandGroup(self, _COMMAND_GROUP, operation_group(_CREATE)) as command_group:
        command_group.command(_CREATE, _CREATE)

    with CommandGroup(self, _COMMAND_GROUP, operation_group(_UPDATE)) as command_group:
        command_group.command(_UPDATE, _UPDATE)

    with CommandGroup(self, _COMMAND_GROUP, operation_group(_VALIDATE)) as command_group:
        command_group.command(_VALIDATE, _VALIDATE)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Prompts the user for missing arguments
"""

from knack.prompting import prompt_choice_list

_PROPERTIES = 'properties'
_VALUE = 'value'
_DISPLAY_NAME = 'displayName'
_NAME = 'name'
_IS_CUSTOM_API = 'isCustomApi'
_CREATED_BY = 'createdBy'


def get_environment(flow_rp):
    """
    Prompt for environment if not provided.
    """
    environments_val = flow_rp.get_environments()
    environments_list = environments_val[_VALUE]
    environments = {
        env[_PROPERTIES][_DISPLAY_NAME]: env[_NAME]
        for env in environments_list
    }

    environment_keys = list(environments.keys())

    sid = prompt_choice_list('Please select an environment:', environment_keys)
    environment = environments[environment_keys[sid]]

    print('Environment selected: {}'.format(environment_keys[sid]))

    return environment


def get_connector_id(powerapps_rp, environment):
    """
    Select connector id if not provided.
    """
    connectors_val = powerapps_rp.get_all_connectors(environment)
    connectors_list = connectors_val[_VALUE]
    custom_connectors = filter(lambda conn: conn[_PROPERTIES][_IS_CUSTOM_API], connectors_list)
    connectors = {
        conn[_PROPERTIES][_DISPLAY_NAME] + ' - ' + conn[_PROPERTIES][_CREATED_BY][_DISPLAY_NAME]: conn[_NAME]
        for conn in custom_connectors
    }

    connectors_keys = list(connectors.keys())

    sid = prompt_choice_list('Please select a connector:', connectors_keys)
    connector_id = connectors[connectors_keys[sid]]

    print('Connector selected: {}'.format(connectors_keys[sid]))

    return connector_id



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Download command.
"""

from paconn import _DOWNLOAD

from paconn.common.util import display
from paconn.settings.util import load_powerapps_and_flow_rp
from paconn.settings.settingsbuilder import SettingsBuilder

import paconn.operations.download


# pylint: disable=too-many-arguments
def download(
        environment,
        connector_id,
        destination,
        powerapps_url,
        powerapps_version,
        settings_file,
        overwrite):
    """
    Download command.
    """
    # Get settings
    settings = SettingsBuilder.get_settings(
        environment=environment,
        settings_file=settings_file,
        api_properties=None,
        api_definition=None,
        icon=None,
        connector_id=connector_id,
        powerapps_url=powerapps_url,
        powerapps_version=powerapps_version)

    powerapps_rp, _ = load_powerapps_and_flow_rp(
        settings=settings,
        command_context=_DOWNLOAD)

    directory = paconn.operations.download.download(
        powerapps_rp=powerapps_rp,
        settings=settings,
        destination=destination,
        overwrite=overwrite)

    display('The connector is downloaded to {}.'.format(directory))



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Method for create/update operation
"""

import io
import json

from paconn.common.util import ensure_file_exists


def validate(powerapps_rp, settings):
    """
    Method for create/update operation
    """

    # Make sure the required files exist
    ensure_file_exists(
        file=settings.api_definition,
        file_type='API Definition')

    # Load swagger definition
    with io.open(settings.api_definition, 'r', encoding='utf-8-sig') as file:
        openapi_definition = json.load(file)

    # Validate Open API Definition
    result = powerapps_rp.validate_connector(
        payload=openapi_definition,
        enable_certification_rules=True)

    # Replace \r\n in the string to newlines
    result = bytes(result, 'utf-8').decode('unicode-escape')
    # Remove quotes at the beginning and end
    result = result.strip('"')

    return result



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Token file manager.
"""

import os
import json

import time

from knack.util import CLIError

from paconn.common.util import get_config_dir

TOKEN_FILE = 'accessTokens.json'

# Token specific variables
_TOKEN_TYPE = 'token_type'
_ACCESS_TOKEN = 'access_token'
_EXPIRES_ON = 'expires_on'
_OID = 'oid'


# Number of seconds to request a login before the token expires
TOKEN_BUFFER_SECONDS = 600


class TokenManager:
    """
    Class to manager login token.
    """
    def __init__(self, token_file=TOKEN_FILE):
        self.token_file = os.path.join(get_config_dir(), token_file)

    def get_credentials(self):
        """
        Returns credential object from token file.
        """
        credentials = self.read()
        token_expired = TokenManager.is_expired(credentials)
        if token_expired:
            raise CLIError('Access token invalid. Please login again.')

        return credentials

    def read(self):
        """
        Reads a login token file.
        """
        creds = []
        if os.path.isfile(self.token_file):
            try:
                with open(self.token_file, 'r') as file:
                    creds = json.load(file)
            except ValueError as exception:
                raise CLIError("Failed to load token files. (Inner Error: {})".format(exception))
        return creds

    def write(self, credentials):
        """
        Writes the login credentials to a token file.
        """
        with os.fdopen(os.open(self.token_file, os.O_RDWR | os.O_CREAT | os.O_TRUNC, 0o600), 'w+') as cred_file:
            cred_file.write(json.dumps(credentials))

    @staticmethod
    def is_expired(credentials):
        """
        Returns true if the token is expired.
        """
        token_expired = _ACCESS_TOKEN not in credentials

        # Check for timeout
        if not token_expired and _EXPIRES_ON in credentials:
            # time.time() returns number of seconds since epoch, 01/01/1970 UTC
            expiration_time = time.time() + TOKEN_BUFFER_SECONDS
            token_expired = credentials[_EXPIRES_ON] < expiration_time

        return token_expired



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Uploads an icon for the custom connector
"""
import os
import mimetypes
from urllib.parse import urlparse, urlunparse
from azure.storage.blob import ContentSettings, BlockBlobService


def upload_icon(sas_url, file_path):
    # Break the SAS URL
    (scheme, netloc, path, params, query, fragment) = urlparse(sas_url)
    # Account is the first part of the netlocation upto the dot
    account_name = netloc[0:netloc.index('.')]

    # The assumption here is that the blob URL will be in the
    # form accountname.blob.core.windows.net or
    # accountname.blob.core.usgovcloudapi.net.
    # Chopping off accountname.blob. to obtain the
    # endpoint suffix.
    endpoint_suffix = netloc.replace(account_name+'.blob.', '')

    # Container name is the path
    container_name = path.strip('/')

    # Create a block blob service
    blockblob_service = BlockBlobService(
        account_name=account_name,
        sas_token=query,
        endpoint_suffix=endpoint_suffix)

    # Get the file name of the icon
    file_name = os.path.basename(file_path)
    # Determine the content type and encoding for the file
    (content_type, content_encoding) = mimetypes.guess_type(file_name)
    content_settings = ContentSettings(
        content_type=content_type,
        content_encoding=content_encoding)

    # Upload the icon
    blockblob_service.create_blob_from_path(
        container_name=container_name,
        blob_name=file_name,
        file_path=file_path,
        content_settings=content_settings)

    # Append the icon name to the path to generate the download link
    path = path + '/' + file_name
    urlparts = (scheme, netloc, path, params, query, fragment)
    sas_download_url = urlunparse(urlparts)

    return sas_download_url



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Represents a settings builder class
"""

from paconn.settings.settings import Settings
from paconn.settings.settingsserializer import SettingsSerializer


# pylint: disable=too-few-public-methods
class SettingsBuilder:
    """
    A builder class to create a Settings object
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def get_settings(
            environment,
            settings_file,
            connector_id,
            powerapps_url,
            powerapps_version,
            api_properties,
            api_definition,
            icon,
            client_id=None,
            tenant=None,
            authority_url=None,
            resource=None):
        """
        Loads settings into a settings object.
        """

        # Load from settings file if it is available
        if settings_file:
            settings = SettingsSerializer.from_json(settings_file)
        else:
            settings = Settings(
                connector_id=connector_id,
                environment=environment,
                powerapps_url=powerapps_url,
                powerapps_api_version=powerapps_version,
                api_properties=api_properties,
                api_definition=api_definition,
                icon=icon,
                client_id=client_id,
                tenant=tenant,
                authority_url=authority_url,
                resource=resource
            )
        return settings

    # pylint: disable=too-many-arguments
    @staticmethod
    def get_authentication_settings(
            settings_file,
            client_id,
            tenant,
            authority_url,
            resource):
        """
        Loads settings into a settings object.
        """

        # Load from settings file if it is available
        if settings_file:
            settings = SettingsSerializer.from_json(settings_file)
        else:
            settings = Settings(
                connector_id=None,
                environment=None,
                powerapps_url=None,
                powerapps_api_version=None,
                api_properties=None,
                api_definition=None,
                icon=None,
                client_id=client_id,
                tenant=tenant,
                authority_url=authority_url,
                resource=resource
            )
        return settings



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
PowerApps RP manager
"""

import json
from urllib.parse import urljoin


class PowerAppsRP:
    """
    PowerAppsRP manager.
    """

    def __init__(self, api_manager):
        self.api_manager = api_manager

    @staticmethod
    def _get_filter_query(environment):
        return {'$filter': 'environment eq \'{}\''.format(environment)}

    def get_connector(self, environment, connector_id):
        """
        Returns API registration JSON for a given connector.
        """
        api = urljoin('apis/', connector_id)

        endpoint = self.api_manager.construct_url(
            path=api,
            query=PowerAppsRP._get_filter_query(environment))

        response = self.api_manager.request(
            verb='GET',
            endpoint=endpoint)

        return response.json()

    def create_connector(self, environment, payload):
        """
        Creates a new custom connector.
        """
        endpoint = self.api_manager.construct_url(
            path='apis',
            query=PowerAppsRP._get_filter_query(environment))

        response = self.api_manager.request(
            verb='POST',
            endpoint=endpoint,
            payload=payload)

        return response.text

    def update_connector(self, environment, connector_id, payload):
        """
        Updates a custom connector.
        """
        api = urljoin('apis/', connector_id)

        endpoint = self.api_manager.construct_url(
            path=api,
            query=PowerAppsRP._get_filter_query(environment))

        response = self.api_manager.request(
            verb='PATCH',
            endpoint=endpoint,
            payload=payload)

        return response.text

    def get_all_connectors(self, environment):
        """
        Returns all connectors.
        """
        endpoint = self.api_manager.construct_url(
            path='apis',
            query=PowerAppsRP._get_filter_query(environment))

        response = self.api_manager.request(
            verb='GET',
            endpoint=endpoint)

        return response.json()

    def validate_connector(self, payload, enable_certification_rules):
        """
        Validates a custom connector.
        """
        api = self.api_manager.add_object_id('validateApiSwagger')

        query = None
        if enable_certification_rules:
            query = {'enableConnectorCertificationRules': 'true'}

        endpoint = self.api_manager.construct_url(
            path=api,
            query=query)

        response = self.api_manager.request(
            verb='POST',
            endpoint=endpoint,
            payload=payload)

        return response.text

    def generate_resource_storage(self, environment):
        """
        Generates a resource storage
        """
        api = self.api_manager.add_object_id('generateResourceStorage')

        endpoint = self.api_manager.construct_url(path=api)

        payload = {'environment': {'name': environment}}

        response = self.api_manager.request(
            verb='POST',
            endpoint=endpoint,
            payload=payload)

        return json.loads(response.text)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Defines argument completer
"""


# pylint: disable=too-few-public-methods
class Completer:
    """
    Argument completer object
    """
    def __init__(self, func):
        self.func = func

    def __call__(self, **kwargs):
        namespace = kwargs['parsed_args']
        prefix = kwargs['prefix']
        cmd = namespace._cmd  # pylint: disable=protected-access
        return self.func(cmd, prefix, namespace)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Utility methods.
"""
import sys
import os
import json

from knack.util import CLIError
from knack.prompting import prompt_y_n


def get_config_dir():
    """
    Returns the user config directory.
    """
    from paconn import __CLI_NAME__
    return os.path.expanduser(os.path.join('~', '.{}'.format(__CLI_NAME__)))


def display(txt):
    """
    Displayed the text to stderr stream.
    """
    print(txt, file=sys.stderr)


def format_json(content, sort_keys=False):
    """
    Format a given dictionary to a json formatted string.
    """
    json_string = json.dumps(
        content,
        sort_keys=sort_keys,
        indent=2,
        separators=(',', ': '))
    return json_string


def ensure_file_exists(file, file_type):
    """
    Check if the given file exists.
    """

    if not file:
        raise CLIError('{} must be specified.'.format(file_type))
    if not os.path.exists(file):
        raise CLIError('File does not exist: {}'.format(file))


def ensure_overwrite(filename):
    overwrite = True
    if os.path.exists(filename):
        msg = '{} file exists. Do you want to overwrite?'.format(filename)
        overwrite = prompt_y_n(msg)

    return overwrite


def write_with_prompt(filename, mode, content, overwrite):
    if not overwrite:
        overwrite = ensure_overwrite(filename)

    if overwrite:
        open(filename, mode=mode).write(content)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Defines constants for JSON keys.
"""

# All
_PROPERTIES = 'properties'
_ICON_URI = 'iconUri'
_CONNECTION_PARAMETERS = 'connectionParameters'

# Save
_API_DEFINITIONS = 'apiDefinitions'
_ORIGINAL_SWAGGER_URL = 'originalSwaggerUrl'
_ICON_BRAND_COLOR = 'iconBrandColor'
_CAPABILITIES = 'capabilities'
_POLICY_TEMPLATE_INSTANCES = 'policyTemplateInstances'

# Create update
_OPEN_API_DEFINITION = 'openApiDefinition'
_ENVIRONMENT = 'environment'
_BACKEND_SERVICE = 'backendService'
_SERVICE_URL = 'serviceUrl'
_SCHEMES = 'schemes'
_HOST = 'host'
_BASE_PATH = 'basePath'
_SHARED_ACCESS_SIGNATURE = 'sharedAccessSignature'
_DESCRIPTION = 'description'
_INFO = 'info'
_TITLE = 'title'
_TOKEN = 'token'
_OAUTH_SETTINGS = 'oAuthSettings'
_CLIENT_SECRET = 'clientSecret'
_DISPLAY_NAME = 'displayName'
_NAME = 'name'



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
CLI parameter definitions
"""

from knack.arguments import ArgumentsContext
from paconn import _LOGIN, _DOWNLOAD, _CREATE, _UPDATE, _VALIDATE

CLIENT_SECRET = 'client_secret'
CLIENT_SECRET_OPTIONS = ['--secret', '-r']
CLIENT_SECRET_HELP = 'The OAuth2 client secret for the connector.'

ENVIRONMENT = 'environment'
ENVIRONMENT_OPTIONS = ['--env', '-e']
ENVIRONMENT_HELP = 'Power Platform environment ID.'

CONNECTOR_ID = 'connector_id'
CONNECTOR_ID_OPTIONS = ['--cid', '-c']
CONNECTOR_ID_HELP = 'The custom connector ID.'

POWERAPPS_URL = 'powerapps_url'
POWERAPPS_URL_OPTIONS = ['--pau', '-u']
POWERAPPS_URL_HELP = 'Power Platform URL.'

POWERAPPS_VERSION = 'powerapps_version'
POWERAPPS_VERSION_OPTIONS = ['--pav', '-v']
POWERAPPS_VERSION_HELP = 'Power Platform api version.'

SETTINGS = 'settings_file'
SETTINGS_OPTIONS = ['--settings', '-s']
SETTINGS_HELP = 'A settings file containing required parameters. When a settings file is specified some commandline parameters are ignored.'  # noqa: E501

API_PROPERTIES = 'api_properties'
API_PROPERTIES_OPTIONS = ['--api-prop', '-p']
API_PROPERTIES_HELP = 'Location of the API properties JSON document.'

API_DEFINITION = 'api_definition'
API_DEFINITION_OPTIONS = ['--api-def', '-d']
API_DEFINITION_HELP = 'Location of the Open API definition JSON document.'

ICON = 'icon'
ICON_OPTIONS = ['--icon', '-i']
ICON_HELP = 'Location for the icon file.'


# pylint: disable=unused-argument
def load_arguments(self, command):
    """
    Load command line arguments
    """
    with ArgumentsContext(self, _LOGIN) as arg_context:
        arg_context.argument(
            'client_id',
            options_list=['--clid', '-i'],
            type=str,
            required=False,
            help='The client ID.')
        arg_context.argument(
            'tenant',
            options_list=['--tenant', '-t'],
            type=str,
            required=False,
            help='The tenant.')
        arg_context.argument(
            'authority_url',
            options_list=['--authority_url', '-a'],
            type=str,
            required=False,
            help='Authority URL for login.')
        arg_context.argument(
            'resource',
            options_list=['--resource', '-r'],
            type=str,
            required=False,
            help='Resource URL for login.')
        arg_context.argument(
            SETTINGS,
            options_list=SETTINGS_OPTIONS,
            type=str,
            required=False,
            help=SETTINGS_HELP)
        arg_context.argument(
            'force',
            options_list=['--force', '-f'],
            type=bool,
            required=False,
            nargs='?',
            default=False,
            const=True,
            help='Override a previous login, if exists.')

    with ArgumentsContext(self, _DOWNLOAD) as arg_context:
        arg_context.argument(
            ENVIRONMENT,
            options_list=ENVIRONMENT_OPTIONS,
            type=str,
            required=False,
            help=ENVIRONMENT_HELP)
        arg_context.argument(
            CONNECTOR_ID,
            options_list=CONNECTOR_ID_OPTIONS,
            type=str,
            required=False,
            help=CONNECTOR_ID_HELP)
        arg_context.argument(
            'destination',
            options_list=['--dest', '-d'],
            type=str,
            required=False,
            help='Destination directory. Non-existent directories will be created.')
        arg_context.argument(
            POWERAPPS_URL,
            options_list=POWERAPPS_URL_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_URL_HELP)
        arg_context.argument(
            POWERAPPS_VERSION,
            options_list=POWERAPPS_VERSION_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_VERSION_HELP)
        arg_context.argument(
            SETTINGS,
            options_list=SETTINGS_OPTIONS,
            type=str,
            required=False,
            help=SETTINGS_HELP)
        arg_context.argument(
            'overwrite',
            options_list=['--overwrite', '-w'],
            type=bool,
            required=False,
            nargs='?',
            default=False,
            const=True,
            help='Overwrite all the existing connector and settings files.')

    with ArgumentsContext(self, _CREATE) as arg_context:
        arg_context.argument(
            ENVIRONMENT,
            options_list=ENVIRONMENT_OPTIONS,
            type=str,
            required=False,
            help=ENVIRONMENT_HELP)
        arg_context.argument(
            API_PROPERTIES,
            options_list=API_PROPERTIES_OPTIONS,
            type=str,
            required=False,
            help=API_PROPERTIES_HELP)
        arg_context.argument(
            API_DEFINITION,
            options_list=API_DEFINITION_OPTIONS,
            type=str,
            required=False,
            help=API_DEFINITION_HELP)
        arg_context.argument(
            ICON,
            options_list=ICON_OPTIONS,
            type=str,
            required=False,
            help=ICON_HELP)
        arg_context.argument(
            POWERAPPS_URL,
            options_list=POWERAPPS_URL_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_URL_HELP)
        arg_context.argument(
            POWERAPPS_VERSION,
            options_list=POWERAPPS_VERSION_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_VERSION_HELP)
        arg_context.argument(
            CLIENT_SECRET,
            options_list=CLIENT_SECRET_OPTIONS,
            type=str,
            required=False,
            help=CLIENT_SECRET_HELP)
        arg_context.argument(
            SETTINGS,
            options_list=SETTINGS_OPTIONS,
            type=str,
            required=False,
            help=SETTINGS_HELP)
        arg_context.argument(
            'overwrite_settings',
            options_list=['--overwrite-settings', '-w'],
            type=bool,
            required=False,
            nargs='?',
            default=False,
            const=True,
            help='Overwrite the existing settings file.')

    with ArgumentsContext(self, _UPDATE) as arg_context:
        arg_context.argument(
            ENVIRONMENT,
            options_list=ENVIRONMENT_OPTIONS,
            type=str,
            required=False,
            help=ENVIRONMENT_HELP)
        arg_context.argument(
            API_PROPERTIES,
            options_list=API_PROPERTIES_OPTIONS,
            type=str,
            required=False,
            help=API_PROPERTIES_HELP)
        arg_context.argument(
            API_DEFINITION,
            options_list=API_DEFINITION_OPTIONS,
            type=str,
            required=False,
            help=API_DEFINITION_HELP)
        arg_context.argument(
            ICON,
            options_list=ICON_OPTIONS,
            type=str,
            required=False,
            help=ICON_HELP)
        arg_context.argument(
            CONNECTOR_ID,
            options_list=CONNECTOR_ID_OPTIONS,
            type=str,
            required=False,
            help=CONNECTOR_ID_HELP)
        arg_context.argument(
            POWERAPPS_URL,
            options_list=POWERAPPS_URL_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_URL_HELP)
        arg_context.argument(
            POWERAPPS_VERSION,
            options_list=POWERAPPS_VERSION_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_VERSION_HELP)
        arg_context.argument(
            CLIENT_SECRET,
            options_list=CLIENT_SECRET_OPTIONS,
            type=str,
            required=False,
            help=CLIENT_SECRET_HELP)
        arg_context.argument(
            SETTINGS,
            options_list=SETTINGS_OPTIONS,
            type=str,
            required=False,
            help=SETTINGS_HELP)

    with ArgumentsContext(self, _VALIDATE) as arg_context:
        arg_context.argument(
            API_DEFINITION,
            options_list=API_DEFINITION_OPTIONS,
            type=str,
            required=False,
            help=API_DEFINITION_HELP)
        arg_context.argument(
            POWERAPPS_URL,
            options_list=POWERAPPS_URL_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_URL_HELP)
        arg_context.argument(
            POWERAPPS_VERSION,
            options_list=POWERAPPS_VERSION_OPTIONS,
            type=str,
            required=False,
            help=POWERAPPS_VERSION_HELP)
        arg_context.argument(
            SETTINGS,
            options_list=SETTINGS_OPTIONS,
            type=str,
            required=False,
            help=SETTINGS_HELP)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
A manager class for the API calls
"""

import json
from urllib.parse import urljoin, urlencode, urlunparse, quote
import requests

from knack.util import CLIError
from knack.log import get_logger

from paconn.common.util import display, format_json
from paconn.authentication.tokenmanager import (
    _ACCESS_TOKEN,
    _TOKEN_TYPE,
    _OID
)

LOGGER = get_logger(__name__)


class APIManager:
    """
    A manager class for API calls
    """
    # pylint: disable=too-many-arguments
    def __init__(self, scheme, region, netlocation, base_path, api_version, credentials=None):
        self.scheme = scheme

        if not region:
            self.netloc = netlocation
        else:
            self.netloc = '{region}.{netlocation}'.format(
                region=region,
                netlocation=netlocation)

        self.base_path = base_path.rstrip('/') + '/'

        self.api_version = api_version

        self.credentials = credentials

    def add_object_id(self, api):
        """
        Add object id to a given api endpoint
        """
        object_id = self.credentials.get(_OID, '')
        path = 'objectIds/{object_id}/{api}'.format(
            object_id=object_id,
            api=api)

        return path

    def construct_url(self, path, params=None, query=None, fragment=None):
        """
        Contruct a URL from a set of parameters
        """
        urlparts = [''] * 6

        urlparts[0] = self.scheme

        urlparts[1] = self.netloc

        urlparts[2] = urljoin(self.base_path, path)

        if params:
            urlparts[3] = urlencode(params)

        all_query = {
            'api-version': self.api_version
        }

        if query:
            all_query.update(query)
        urlparts[4] = urlencode(all_query, quote_via=quote)

        if fragment:
            urlparts[5] = fragment

        endpoint = urlunparse(urlparts)

        return endpoint

    def request(self, verb, endpoint, headers=None, payload=None):
        """
        Send a request to the given url
        """
        all_headers = {}
        if self.credentials:
            token_type = self.credentials[_TOKEN_TYPE]
            token = self.credentials[_ACCESS_TOKEN]
            all_headers = {
                'Authorization': '{token_type} {token}'.format(
                    token_type=token_type,
                    token=token)
            }
        if headers:
            all_headers.update(headers)

        response = requests.request(
            verb,
            endpoint,
            headers=all_headers,
            json=payload)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as exception:
            exception_str = str(exception)
            response_content = json.loads(response.content)
            response_content = format_json(response_content)
            if payload:
                LOGGER.debug('PAYLOAD')
                LOGGER.debug(payload)
            LOGGER.debug('RESPONSE')
            LOGGER.debug(response_content)
            display(response_content)
            raise CLIError(exception_str)

        return response



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Represents a settings object consructed from settings.json
"""


# pylint: disable=too-few-public-methods
class Settings:
    """
    Represents a settings object consructed from settings.json
    """
    # pylint: disable=too-many-arguments
    def __init__(
            self,
            connector_id,
            environment,
            api_properties,
            api_definition,
            icon,
            powerapps_url,
            powerapps_api_version,
            flow_url=None,
            flow_api_version=None,
            client_id=None,
            tenant=None,
            authority_url=None,
            resource=None):

        # connector specific settings
        self.connector_id = connector_id
        self.environment = environment

        # Files
        self.api_properties = api_properties
        self.api_definition = api_definition
        self.icon = icon

        # PowerApps RP settings
        self.powerapps_url = powerapps_url or 'https://api.powerapps.com'
        self.powerapps_api_version = powerapps_api_version or '2016-11-01'
        self.powerapps_base_path = 'providers/Microsoft.PowerApps'         # Constant

        # Flow RP Settings
        self.flow_url = flow_url or 'https://api.flow.microsoft.com'
        self.flow_api_version = flow_api_version or '2016-11-01'
        self.flow_base_path = 'providers/Microsoft.ProcessSimple'          # Constant

        self.client_id = client_id or '04b07795-8ddb-461a-bbee-02f9e1bf7b46'
        self.tenant = tenant or 'common'
        self.authority_url = authority_url or 'https://login.microsoftonline.com/'
        self.resource = resource or 'https://management.core.windows.net/'



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""Microsoft Power Platform Connectors CLI package that can be installed using setuptools"""

import os
from setuptools import setup

__VERSION__ = '0.0.13'


def read(fname):
    """Local read helper function for long documentation"""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='paconn',
    version=__VERSION__,
    description='Microsoft Power Platform Connectors CLI',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    author='Microsoft Corporation',
    author_email='connectors@microsoft.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='azure, powerapps, flow, power platform, connectors',
    python_requires='>=3.5,<4',
    packages=[
        'paconn',
        'paconn.apimanager',
        'paconn.authentication',
        'paconn.commands',
        'paconn.common',
        'paconn.config',
        'paconn.operations',
        'paconn.settings'
    ],
    install_requires=[
        'docutils',
        'flake8',
        'future',
        'knack~=0.5.1',
        'pytest',
        'pytest-xdist',
        'virtualenv',
        'requests',
        'adal',
        'msrestazure',
        'azure-storage-blob>=2.1,<12.0'
    ],
    extras_require={
        ":python_version<'3.0'": ['pylint~=1.9.2'],
        ":python_version>='3.0'": ['pylint~=2.0.0']
    },
    package_data={'paconn.config': ['*.*']},
    include_package_data=True,
    entry_points={
        'console_scripts': ['paconn=paconn.__main__:main']
    }
)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
Initializer
"""

__VERSION__ = '0.0.13'
__CLI_NAME__ = 'paconn'

# Commands
_COMMAND_GROUP = ''
_LOGIN = 'login'
_DOWNLOAD = 'download'
_CREATE = 'create'
_UPDATE = 'update'
_VALIDATE = 'validate'



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
A builder class to create an API Manager object
"""

from urllib.parse import urlparse
from paconn.apimanager.apimanager import APIManager


# pylint: disable=too-few-public-methods
class APIManagerBuilder:
    """
    A builder class to create an API Manager object from an url
    """
    @staticmethod
    def get_from_url(url, base_path, api_version, credentials):
        """
        Creates an APIManager object from given URL, Base Path and credentials
        """
        # pylint: disable=unused-variable
        (scheme, netloc, path, params, query, fragment) = urlparse(url)

        return APIManager(
            scheme=scheme,
            region=None,
            netlocation=netloc,
            base_path=base_path,
            api_version=api_version,
            credentials=credentials)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""
User profile management class.`
"""
import adal
from urllib.parse import urljoin
# AADTokenCredentials for multi-factor authentication
from msrestazure.azure_active_directory import AADTokenCredentials


class Profile:
    """
    A Class representing user profile.
    """

    def __init__(self, client_id, tenant, resource, authority_url):
        self.client_id = client_id
        self.tenant = tenant
        self.resource = resource
        self.authority_url = authority_url

    def _get_authentication_context(self):
        auth_url = urljoin(self.authority_url, self.tenant)

        return adal.AuthenticationContext(
            authority=auth_url,
            api_version=None)

    def authenticate_device_code(self):
        """
        Authenticate the end-user using device auth.
        """
        context = self._get_authentication_context()

        code = context.acquire_user_code(
            resource=self.resource,
            client_id=self.client_id)

        print(code['message'])

        mgmt_token = context.acquire_token_with_device_code(
            resource=self.resource,
            user_code_info=code,
            client_id=self.client_id)

        credentials = AADTokenCredentials(
            token=mgmt_token,
            client_id=self.client_id)

        return credentials.token



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Utility for loading settings.
"""

from paconn import _UPDATE, _DOWNLOAD, _VALIDATE
from paconn.common.util import write_with_prompt
from paconn.authentication.tokenmanager import TokenManager
from paconn.apimanager.powerappsrpbuilder import PowerAppsRPBuilder
from paconn.apimanager.flowrpbuilder import FlowRPBuilder
from paconn.common.prompts import get_environment, get_connector_id
from paconn.settings.settingsserializer import SettingsSerializer

# Setting file name
SETTINGS_FILE = 'settings.json'


def prompt_for_environment(settings, flow_rp):
    # Select environment if not provided
    if not hasattr(settings, 'environment') or not settings.environment:
        settings.environment = get_environment(
            flow_rp=flow_rp)


def prompt_for_connector_id(settings, powerapps_rp):
    # Select connector id if not provided
    if (not hasattr(settings, 'connector_id') or not settings.connector_id):
        settings.connector_id = get_connector_id(
            powerapps_rp=powerapps_rp,
            environment=settings.environment)


def load_powerapps_and_flow_rp(settings, command_context):

    # Get credentials
    credentials = TokenManager().get_credentials()

    # Get powerapps rp
    powerapps_rp = PowerAppsRPBuilder.get_from_settings(
        credentials=credentials,
        settings=settings)

    # Get flow rp
    flow_rp = FlowRPBuilder.get_from_settings(
        credentials=credentials,
        settings=settings)

    # If the file names are missing for the download command
    # use default names
    if command_context is _DOWNLOAD:
        settings.api_properties = settings.api_properties or 'apiProperties.json'
        settings.api_definition = settings.api_definition or 'apiDefinition.swagger.json'
        settings.icon = settings.icon or 'icon.png'

    # Prompt for environment for any
    # operation other than `validate`
    if command_context is not _VALIDATE:
        prompt_for_environment(
            settings=settings,
            flow_rp=flow_rp)

    # Prompt for connector id for
    # operation `update` and `download`
    if command_context is _UPDATE or command_context is _DOWNLOAD:
        prompt_for_connector_id(
            settings=settings,
            powerapps_rp=powerapps_rp)

    return powerapps_rp, flow_rp


def write_settings(settings, overwrite):
    filename = SETTINGS_FILE
    settings_json = SettingsSerializer.to_json_string(settings)
    write_with_prompt(
        filename=filename,
        mode='w',
        content=settings_json,
        overwrite=overwrite)



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Represents a settings object consructed from settings.json
"""

import json
from paconn.common.util import format_json
from paconn.settings.settings import Settings

# Connector sepecific settings
_CONNECTOR_ID = 'connectorId'
_ENVIRONMENT = 'environment'

# PowerApps RP settings
_POWERAPPS_URL = 'powerAppsUrl'
_POWERAPPS_API_VERSION = 'powerAppsApiVersion'

# Flow RP Settings
_FLOW_URL = 'flowUrl'
_FLOW_API_VERSION = 'flowApiVersion'

# Files
_API_PROPERTIES = 'apiProperties'
_API_DEFINITION = 'apiDefinition'
_ICON = 'icon'

# Authentication settings
_CLIENT_ID = 'clientId'
_TENANT = 'tenant'
_AUTHORITY_URL = 'authorityUrl'
_RESOURCE = 'resource'


# pylint: disable=too-few-public-methods
class SettingsSerializer:
    """
    Serializes and deserializes a settings object
    """
    @staticmethod
    def to_json_string(settings):
        """
        Serializes a settings object into string
        """
        settings_dict = SettingsSerializer.serialize(settings)
        json_str = format_json(settings_dict)
        return json_str

    @staticmethod
    def to_json(settings, filename):
        """
        Serializes a settings object into the settings.json
        """
        json_str = SettingsSerializer.to_json_string(settings)
        open(filename, 'w').write(json_str)

    @staticmethod
    def from_json(filename):
        """
        Deserializes a settings object from the settings.json file
        """
        with open(filename, 'r') as file:
            settings_dict = json.load(file)  # pylint: disable=attribute-defined-outside-init
        settings = SettingsSerializer.deserialize(settings_dict)
        return settings

    @staticmethod
    def serialize(settings):
        """
        Serializes a settings object into a dict
        """
        settings_dict = {
            # Connector sepecific settings
            _CONNECTOR_ID: settings.connector_id,
            _ENVIRONMENT: settings.environment,

            # Files
            _API_PROPERTIES: settings.api_properties,
            _API_DEFINITION: settings.api_definition,
            _ICON: settings.icon,

            # PowerApps RP settings
            _POWERAPPS_URL: settings.powerapps_url,
            _POWERAPPS_API_VERSION: settings.powerapps_api_version
            # powerapps_base_path IGNORED

            # Flow RP Settings - DO NOT WRITE
            # flow_url IGNORED
            # flow_api_version IGNORED
            # flow_base_path IGNORED

            # Authentication Settings - DO NOT WRITE
            # client_id IGNORED
            # tenant IGNORED
            # authority_url IGNORED
            # resource IGNORED
        }
        return settings_dict

    @staticmethod
    def deserialize(settings_dict):
        """
        Deserializes a dictionary to a settings object
        """
        settings = Settings(
            # Connector sepecific settings
            connector_id=settings_dict.get(_CONNECTOR_ID, None),
            environment=settings_dict.get(_ENVIRONMENT, None),

            # Files
            api_properties=settings_dict.get(_API_PROPERTIES, None),
            api_definition=settings_dict.get(_API_DEFINITION, None),
            icon=settings_dict.get(_ICON, None),

            # PowerApps RP settings
            powerapps_url=settings_dict.get(_POWERAPPS_URL, None),
            powerapps_api_version=settings_dict.get(_POWERAPPS_API_VERSION, None),
            # powerapps_base_path IGNORED

            # Flow RP Settings
            flow_url=settings_dict.get(_FLOW_URL, None),
            flow_api_version=settings_dict.get(_FLOW_API_VERSION, None),
            # flow_base_path IGNORED

            # Authentication Settings
            client_id=settings_dict.get(_CLIENT_ID, None),
            tenant=settings_dict.get(_TENANT, None),
            authority_url=settings_dict.get(_AUTHORITY_URL, None),
            resource=settings_dict.get(_RESOURCE, None)
        )
        return settings



# -----------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""
Login command.
"""

from paconn.authentication.auth import get_authentication
from paconn.common.util import display
from paconn.settings.settingsbuilder import SettingsBuilder


def login(client_id, tenant, authority_url, resource, settings_file, force):
    """
    Login command.
    """
    # Get settings
    settings = SettingsBuilder.get_authentication_settings(
        settings_file=settings_file,
        client_id=client_id,
        tenant=tenant,
        authority_url=authority_url,
        resource=resource)

    get_authentication(
        settings=settings,
        force_authenticate=force)
    display('Login successful.')
